from collections import defaultdict
from datetime import datetime, timedelta
import logging
import multiprocessing
import pathlib
import tempfile
import warnings
from typing import Union

from appdirs import user_data_dir
import boto3
from botocore import UNSIGNED
from botocore.config import Config
import geopandas as gpd
from netCDF4 import Dataset
import numpy as np
from psutil import cpu_count
import pytz
from scipy.spatial import cKDTree
from shapely import ops
from shapely.geometry import (
    LinearRing, Point, MultiPoint, LineString)
# from tqdm import tqdm
import wget

from pyschism.mesh import Hgrid, Gr3
from pyschism.forcing.hydrology.base import Hydrology, Sources, Sinks


DATADIR = pathlib.Path(user_data_dir()) / 'nwm'
DATADIR.mkdir(exist_ok=True, parents=True)
NWM_FILE = DATADIR / 'nwm_v12.gdb.zip'

_logger = logging.getLogger(__name__)


class NWMGeoDataFrame:

    def __get__(self, obj, val):
        gdf = obj.__dict__.get('gdf')
        if gdf is None:
            bbox = obj._hgrid.bbox
            gdf = gpd.read_file(
                NWM_FILE,
                bbox=(bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax),
                # bbox=(-75.889435, 38.895308, -74.604034, 39.477546),  # Delaware Bay, debug  # noqa: E501
                )
            obj.__dict__['gdf'] = gdf
        return gdf

    def __delete__(self, obj):
        del obj.__dict__['gdf']


class NWMElementPairings:

    _gdf = NWMGeoDataFrame()

    def __init__(self, hgrid):
        _logger.info('Initiliaze NWMElementPairings')
        _logger.debug('This debug message should also appear.')
        self._hgrid = hgrid

        # An STR-Index returns the reaches that are near the boundaries of the
        # mesh. This subsamples the NWM network, but also is not the exact
        # result. This is used to speed-up computations by filtering the input
        # data.
        nwm_r_index = self._gdf.sindex

        # The r-index is used to find intersections between mesh boundary edges
        # and NWM reaches (approximate results)
        possible_indexes = set()
        for edge in hgrid.hull.edges().itertuples():
            for index in list(nwm_r_index.intersection(edge.geometry.bounds)):
                possible_indexes.add(index)
        possible_matches = self._gdf.iloc[list(possible_indexes)]
        del possible_indexes
        del nwm_r_index

        # The hull rings itersections is used to find the exact NWM reaches
        # that intersect the mesh's hull.
        exact_indexes = set()
        for pm in possible_matches.itertuples():
            if hgrid.hull.rings().geometry.intersects(pm.geometry).any():
                exact_indexes.add(pm.Index)
        reaches = self._gdf.iloc[list(exact_indexes)]

        # release some memory
        del possible_matches
        del exact_indexes
        del self._gdf

        # Pair each reach with corresponding element.
        # 1) Find reach-hull intersections.
        data = []
        intersections: gpd.GeoDataFrame
        for i, reach in enumerate(reaches.itertuples()):
            for ring in hgrid.hull.rings().itertuples():
                if ring.geometry.intersects(reach.geometry):
                    data.append({
                        "geometry": ring.geometry.intersection(reach.geometry),
                        "reachIndex": i})
                    break

        intersections = gpd.GeoDataFrame(data, crs=hgrid.crs)
        del data

        # 2) Generate element centroid KDTree
        centroids = []
        for element in hgrid.elements().values():
            cent = LinearRing(hgrid.nodes.coord()[list(
                map(hgrid.nodes.get_index_by_id, element))]
            ).centroid
            centroids.append((cent.x, cent.y))
        tree = cKDTree(centroids)
        del centroids

        # 3) Match reach/boundary intersection to nearest element centroid
        element_index = {}
        for inters in intersections.itertuples():
            geom = inters.geometry
            if isinstance(geom, Point):
                geom = MultiPoint(geom.coords)
            element_index[inters.reachIndex] = []
            for point in geom:
                _, idx = tree.query(point)
                element_index[inters.reachIndex].append(idx)
        del tree

        elements = hgrid.elements.geodataframe()
        del self._hgrid  # release

        sources = defaultdict(set)
        sinks = defaultdict(set)
        for reach_index, paired_elements_idxs in element_index.items():
            reach = reaches.iloc[reach_index]
            point_of_intersection = intersections.loc[
                intersections["reachIndex"] == reach_index]
            for element_idx in paired_elements_idxs:
                element = elements.iloc[element_idx]
                if not isinstance(reach.geometry, LineString):
                    geom = ops.linemerge(reach.geometry)
                else:
                    geom = reach.geometry
                for segment in map(LineString, zip(geom.coords[:-1],
                                                   geom.coords[1:])):
                    if segment.intersects(
                            point_of_intersection.iloc[0].geometry.buffer(
                                np.finfo(np.float32).eps)):
                        downstream = segment.coords[-1]
                        if element.geometry.contains(Point(downstream)):
                            sources[element.id].add(reach.featureID)
                        else:
                            sinks[element.id].add(reach.featureID)
                        break
        self._sources = sources
        self._sinks = sinks

    @property
    def sources(self):
        return self._sources

    @property
    def sinks(self):
        return self._sinks

    @property
    def _hgrid(self):
        return self.__hgrid

    @_hgrid.setter
    def _hgrid(self, hgrid: Gr3):
        hgrid = Hgrid(**hgrid.to_dict())
        hgrid.transform_to(gpd.read_file(NWM_FILE, rows=1).crs)
        self.__hgrid = hgrid

    @_hgrid.deleter
    def _hgrid(self):
        del self.__hgrid


class NWMDataGetter:

    def __init__(
            self,
            pairings: NWMElementPairings,
            start_date: datetime,
            rnday: timedelta
    ):
        self._pairings = pairings
        # naïve-datetime
        if start_date.tzinfo is None or \
                start_date.tzinfo.utcoffset(start_date) is None:
            start_date = pytz.timezone('utm').localize(start_date)
        self._start_date = start_date
        self._rnday = rnday


def streamflow_lookup(file, pairings):
    nc = Dataset(file)
    sources = []
    # TODO: read scaling factor directly from netcdf file?
    for element_id, features in pairings.sources.items():
        sources.append(0.01*np.sum(nc['streamflow'][
            np.where(np.isin(nc['feature_id'], list(features)))]))
    sinks = []
    for element_id, features in pairings.sinks.items():
        sinks.append(-0.01*np.sum(nc['streamflow'][
            np.where(np.isin(nc['feature_id'], list(features)))]))
    return (sources, sinks)


def pivot_time(input_datetime=None, period=6):
    if input_datetime is None:
        input_datetime = pytz.timezone('UTC').localize(datetime.utcnow())
    current_cycle = int(period * np.floor(input_datetime.hour / period))
    return pytz.timezone('UTC').localize(
        datetime(input_datetime.year, input_datetime.month,
                 input_datetime.day, current_cycle))


class AWSDataInventory:

    def __init__(
            self,
            start_date: datetime,
            rnday: Union[int, float, timedelta] = None,
            product='medium_range_mem1',
            verbose=False,
            fallback=True,
    ):
        """This will download the latest National Water Model data.

        NetCDF files are saved to the system's temporary directory.
        The AWS data goes back 30 days. For requesting hindcast data from
        before we need a different data source
        """
        _logger.info('Initialize AWSDataInventory')
        self.start_date = start_date
        self.rnday = rnday

        requested_time = pivot_time(start_date) - timedelta(days=1)

        while requested_time <= self.pivot_time:
            self._get_data_from_bucket(requested_time, product, fallback)
            requested_time += timedelta(days=1)

    @property
    def Bucket(self):
        return 'noaa-nwm-pds'

    @property
    def pivot_time(self):
        return pivot_time()

    @property
    def s3(self):
        try:
            return self._s3
        except AttributeError:
            self._s3 = boto3.client(
                's3', config=Config(signature_version=UNSIGNED))
            return self._s3

    @property
    def tmpdir(self):
        try:
            return self._tmpdir
        except AttributeError:
            self._tmpdir = tempfile.TemporaryDirectory()
            return self._tmpdir

    def _get_data_from_bucket(self, requested_time, product, fallback):
        end_date = self.start_date + self.rnday
        nwm_time = requested_time.strftime('%Y%m%d')
        res = self.s3.list_objects_v2(Bucket=self.Bucket, Delimiter='/',
                                      Prefix=f'nwm.{nwm_time}/{product}/')

        if requested_time == pivot_time() and res['KeyCount'] == 0:
            if fallback is True:
                warnings.warn('NWM data is late, defaulting to previous.')
                requested_time -= timedelta(days=1)
                nwm_time = requested_time.strftime('%Y%m%d')
                res = self.s3.list_objects_v2(
                    Bucket=self.Bucket,
                    Delimiter='/',
                    Prefix=f'nwm.{nwm_time}/{product}/'
                )
            else:
                raise IOError(
                    'NWM data is "late", no NWM-data for current cycle.')
        for data in res['Contents']:
            if 'channel' in data['Key']:
                filename = pathlib.Path(
                    self.tmpdir.name) / data['Key'].split('/')[-1]
                with open(filename, 'wb') as f:
                    self.s3.download_fileobj(self.Bucket, data['Key'], f)
                nc = Dataset(filename)
                # TODO: Will crash if the user localized with timedelta object.
                if end_date.tzinfo.localize(
                    datetime.strptime(nc.model_output_valid_time,
                                      '%Y-%m-%d_%H:%M:%S')) > end_date:
                    break

    def __call__(self, pairings: NWMElementPairings, h0=1e-1, nprocs=-1):
        _logger.info('Will pair NWM data to elements...')
        from time import time as _time
        start = _time()
        files = sorted(list(pathlib.Path(self.tmpdir.name).glob('*')))
        with multiprocessing.Pool(
                processes=cpu_count() if nprocs == -1 else nprocs
        ) as pool:
            res = pool.starmap(
                streamflow_lookup, [(file, pairings) for file in files])
        pool.join()
        sources = Sources()
        sinks = Sinks()
        for i, file in enumerate(files):
            nc = Dataset(file)
            time = pytz.timezone('UTC').localize(
                datetime.strptime(
                    nc.model_output_valid_time,
                    "%Y-%m-%d_%H:%M:%S"))
            # TODO: This is slow, it might change if add_data is vectorized.
            for j, element_id in enumerate(pairings.sources.keys()):
                sources.add_data(time, element_id, res[i][0][j], -9999, 0.)
            for k, element_id in enumerate(pairings.sinks.keys()):
                sinks.add_data(time, element_id, res[i][1][k])
        _logger.info(f'Done pairing, took {_time()-start} seconds...')
        return sources, sinks


class AWSDataGetter(NWMDataGetter):

    def __call__(self, product: str = 'medium_range_mem1',
                 verbose: bool = False, h0=1e-1, nprocs=-1):
        """
        We just picked up a server_config. This part must be slurm-aware and
        potentially
        """
        return AWSDataInventory(
            start_date=self._start_date,
            rnday=self._rnday,
            product=product,
            verbose=False
        )(self._pairings, h0, nprocs)


class FTPDataGetter(NWMDataGetter):

    def __call__(self):
        raise NotImplementedError('FTPDataGetter')


class NationalWaterModel(Hydrology):

    def __init__(self):
        self._nwm_file = NWM_FILE
        if not self._nwm_file.exists():
            _logger.warning(
                "Downloading National Water Model stream network file to "
                "the pyschism cache...")
            wget.download(
                "https://www.dropbox.com/s/3w8i46uumbcs49v/nwm_v12.gdb.zip"
                "?dl=1", out=str(self._nwm_file))
        super().__init__()

    def __call__(self, model_driver, nramp_ss: bool = False, dramp_ss=None,
                 h0=1e-1, nprocs=-1):
        """Initializes the NWM data.
        Used by :class:`pyschism.driver.ModelDriver`

        Will pick the "best" data source based on start date and rnday.
        The are fringe cases not yet covered, for example when the data spans
        more than 1 data source.
        """
        super().__call__(model_driver)
        _logger.info('NationalWaterModel.__call__')
        pairings = NWMElementPairings(model_driver.model_domain.hgrid)
        start_date = model_driver.param.opt.start_date
        rnday = model_driver.param.core.rnday
        if start_date >= pivot_time() - timedelta(days=30):
            _logger.info('Fetching NWM data.')
            AWSData = AWSDataGetter(pairings, start_date, rnday)
            sources, sinks = AWSData(h0=h0, nprocs=nprocs)

        else:
            raise NotImplementedError(
                'start_date is less than pivot_time - 30 days ')

        # turn 'on' the source/sink system in SCHISM.
        model_driver.param.opt.if_source = 1

        # set the ramps if applicable, no ramp by default.
        if int(nramp_ss) != 0:
            # nramp_ss = 1 # needed if if_source=1; ramp-up flag for
            # source/sinks
            model_driver.param.opt.nramp_ss = nramp_ss
            # dramp_ss = 2 # needed if if_source=1; ramp-up period in days
        if int(nramp_ss) != 0 and dramp_ss is not None:
            model_driver.param.opt.dramp_ss = dramp_ss

    @staticmethod
    def from_files(msource, vsource, vsink):
        raise NotImplementedError
        # TODO: File parsers.
        nwm = NationalWaterModel()
        nwm._msource = None
        nwm._vsource = None
        nwm._vsink = None
        return nwm



#! /usr/bin/env python
import tempfile
import pathlib
from unittest.mock import patch
from pyschism.mesh import Hgrid
from pyschism.mesh.friction import Fgrid
import unittest


class HgridTestCase(unittest.TestCase):

    def setUp(self):
        self.nodes = {
            '1': ((0., 0.), -5.),
            '2': ((.5, 0.), -4.),
            '3': ((1., 0.), -3.),
            '4': ((1., 1.), -2.),
            '5': ((0., 1.), -1.),
            '6': ((.5, 1.5), 0.),
            '7': ((.33, .33), 1.),
            '8': ((.66, .33), 2.),
            '9': ((.5, .66), 3.),
            '10': ((-1., 1.), 4.),
            '11': ((-1., 0.), 5.),
            }
        self.elements = {
            '1': ['5', '7', '9'],
            '2': ['1', '2', '7'],
            '3': ['2', '3', '8'],
            '4': ['8', '7', '2'],
            '5': ['3', '4', '8'],
            '6': ['4', '9', '8'],
            '7': ['4', '6', '5'],
            '8': ['5', '10', '11', '1'],
            '9': ['9', '4', '5'],
            '10': ['5', '1', '7']
            }

        self.boundaries = dict()

        self.boundaries[None] = {  # "open" boundaries
                0: {'indexes': ['10', '11', '1', '2']},
                1: {'indexes': ['2', '3', '4']}
        }

        self.boundaries[0] = {  # "land" boundaries
            0: {'indexes': ['4', '6']},
            1: {'indexes': ['6',  '5', '10']}
        }

        self.boundaries[1] = {  # "interior" boundary
            0: {'indexes': ['7', '8', '9', '7']}
        }

        self.grd = {
            'nodes': self.nodes,
            'elements': self.elements,
            'boundaries': self.boundaries,
            'description': 'gr3_unittest'
        }

    def test_triangles_only(self):
        self.assertIsInstance(
            Hgrid(
                self.nodes,
                {id: geom for geom in self.elements.values() if len(geom) == 3}
                ),
            Hgrid
            )

    def test_quads_only(self):
        self.assertIsInstance(
            Hgrid(
                self.nodes,
                {id: geom for geom in self.elements.values() if len(geom) == 4}
                ),
            Hgrid
            )

    def test_hybrid(self):
        self.assertIsInstance(Hgrid(self.nodes, self.elements), Hgrid)

    def test_open(self):
        tmpfile = tempfile.NamedTemporaryFile()
        with open(tmpfile.name, 'w') as f:
            f.write('\n')
            f.write(f'{len(self.elements):d} ')
            f.write(f'{len(self.nodes):d}\n')
            for id, ((x, y), z) in self.nodes.items():
                f.write(f"{id} ")
                f.write(f"{x} ")
                f.write(f"{y} ")
                f.write(f"{z}\n")
            for id, geom in self.elements.items():
                f.write(f"{id} ")
                f.write(f"{len(geom)} ")
                for idx in geom:
                    f.write(f"{idx} ")
                f.write(f"\n")
        self.assertIsInstance(Hgrid.open(tmpfile.name), Hgrid)

    @patch('matplotlib.pyplot.show')
    def test_make_plot(self, mock):
        h = Hgrid(self.nodes, self.elements)
        h.make_plot(
            show=True,
            extent=[0, 1, 0, 1],
            title='test',
            cbar_label='elevation [m]',
            vmax=0.
            )
        self.assertIsInstance(h, Hgrid)

    def test_plot_boundary(self):
        h = Hgrid(**self.grd)
        h.plot_boundary(None, 0)
        self.assertIsInstance(h, Hgrid)

    def test_make_plot_wet_only(self):
        nodes = {
            0: ((0., 0.), 0.),
            1: ((1., 0.), -1.),
            2: ((1., 1.), -2.),
            3: ((0., 1.), -3.),
            4: ((0.5, 1.5), -4.),
        }
        elements = {
            0: [2, 4, 3],
            1: [0, 1, 2, 3],
        }
        h = Hgrid(nodes, elements)
        h.make_plot()
        self.assertIsInstance(h, Hgrid)

    def test_write(self):
        h = Hgrid(self.nodes, self.elements)
        tmpdir = tempfile.TemporaryDirectory()
        h.write(pathlib.Path(tmpdir.name) / 'test_hgrid.gr3')
        self.assertIsInstance(h, Hgrid)

    def test_add_custom_boundary_custom(self):
        h = Hgrid(self.nodes, self.elements)
        h.add_boundary_type('ibtype')
        indexes = [('2', '7'), ('3', '8'), ('4', '9')]
        props = {'flow': [1, 2, 3]}
        h.set_boundary_data('ibtype', 0, indexes, properties=props)

    def test_add_boundary_custom_raise(self):
        h = Hgrid(self.nodes, self.elements)
        h.add_boundary_type('ibtype')
        indexes = [('2', '7'), ('3', '10000'), ('4', '9')]
        props = {'flow': [1, 2, 3]}
        self.assertRaises(
            AssertionError,
            h.set_boundary_data,
            'ibtype',
            0,
            indexes,
            **props
        )

    def test_add_boundary(self):
        h = Hgrid(self.nodes, self.elements)
        h.add_boundary_type('ibtype')
        indexes = ['1']
        h.set_boundary_data('ibtype', 0, indexes)

    def test_add_boundary_raise(self):
        h = Hgrid(self.nodes, self.elements)
        h.add_boundary_type('ibtype')
        indexes = ['10000']
        self.assertRaises(
            AssertionError,
            h.set_boundary_data,
            'ibtype',
            0,
            indexes,
        )

    def test_delete_boundary_type(self):
        msh = Hgrid(self.nodes, self.elements)
        msh.delete_boundary_type(None)

    def test_delete_boundary_data(self):
        msh = Hgrid(self.nodes, self.elements, boundaries=self.boundaries)
        msh.delete_boundary_data(None, 0)

    def test_add_existing_boundary_type_raises(self):
        msh = Hgrid(self.nodes, self.elements)
        self.assertRaises(Exception, msh.add_boundary_type, None)

    def test_plot_boundaries(self):
        h = Hgrid(self.nodes, self.elements, self.boundaries)
        h.plot_boundaries()

    def test_triplot(self):
        h = Hgrid(self.nodes, self.elements, self.boundaries)
        h.triplot()

    def test_make_plot_flat_domain(self):
        nodes = {id: (coord, 0.) for id, (coord, _) in self.nodes.items()}
        h = Hgrid(nodes, self.elements, self.boundaries)
        h.make_plot()

    def test_write_boundaries(self):
        tmpdir = tempfile.TemporaryDirectory()
        shp = pathlib.Path(tmpdir.name).absolute()
        msh = Hgrid(
            self.nodes,
            self.elements,
            crs="EPSG:3395",
            boundaries=self.boundaries)
        msh.write_boundaries(shp, overwrite=True)

    def test_write_boundaries_raises(self):
        tmpdir = tempfile.TemporaryDirectory()
        shp = pathlib.Path(tmpdir.name).absolute()
        msh = Hgrid(
            self.nodes,
            self.elements,
            crs="EPSG:3395",
            boundaries=self.boundaries)
        msh.logger.debug('coverage')
        self.assertRaises(IOError, msh.write_boundaries, shp)

    def test_sms2dm(self):
        self.boundaries[None][0].update({'properties': {}})
        msh = Hgrid(
            self.nodes,
            self.elements,
            crs="EPSG:3395",
            boundaries=self.boundaries)
        self.assertIsInstance(msh.sms2dm, str)

    def test_nan_boundaries_raises(self):
        self.boundaries[None][0].update({'properties': {}})
        import numpy as np
        self.nodes['1'] = ((0., 0.), np.nan)
        msh = Hgrid(
            self.nodes,
            self.elements,
            crs="EPSG:3395",
            boundaries=self.boundaries)
        self.assertRaises(Exception, msh.generate_boundaries)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import pathlib
import setuptools  # type: ignore[import]
import subprocess
import sys

subprocess.check_call(
    [sys.executable, '-m', 'pip', 'install', '--upgrade', 'pip'])

subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'wheel'])

try:
    from dunamai import Version
except ImportError:
    subprocess.check_call(
            [sys.executable, '-m', 'pip', 'install', 'dunamai']
    )
    from dunamai import Version  # type: ignore[import]

try:
    version = Version.from_any_vcs().serialize()
except RuntimeError:
    version = '0.0.0'


parent = pathlib.Path(__file__).parent.absolute()
conf = setuptools.config.read_configuration(parent / 'setup.cfg')
meta = conf['metadata']
setuptools.setup(
    name=meta['name'],
    version=version,
    author=meta['author'],
    author_email=meta['author_email'],
    description=meta['description'],
    long_description=meta['long_description'],
    long_description_content_type="text/markdown",
    url=meta['url'],
    packages=setuptools.find_packages(),
    python_requires='>=3.6, <3.9',
    setup_requires=['wheel', 'setuptools_scm', 'setuptools>=41.2'],
    include_package_data=True,
    extras_require={'dev': ['coverage', 'flake8', 'nose']},
    install_requires=[
        'matplotlib',
        'netcdf-flattener>=1.2.0',
        'netCDF4',
        'pyproj',
        'shapely',
        'fiona',
        'f90nml',
        'psutil',
        'scipy',
        'wget',
        'appdirs',
        'cf-python',
        'sqlalchemy',
        'geopandas',
        'pyugrid',
        'pytz',
        'boto3',
        'rtree'
    ],
    entry_points={'console_scripts': ['pyschism = pyschism.__main__:main']},
    tests_require=['nose'],
    test_suite='nose.collector',
)



from pyschism.forcing.atmosphere import nws
from pyschism.forcing.atmosphere.nws.nws import NWS
from pyschism.forcing.atmosphere.nws.nws2 import NWS2
from pyschism.forcing.atmosphere import gfs
from pyschism.forcing.atmosphere.gfs import GlobalForecastSystem


__all__ = [
    'nws',
    "NWS",
    "NWS2",
    'gfs',
    'GlobalForecastSystem'
]



import pathlib

from netCDF4 import Dataset
import numpy as np
from scipy.interpolate import griddata

from pyschism.forcing.tides.base import TidalDataProvider

# https://icdc.cen.uni-hamburg.de/en/hamtide.html
base_url = 'https://icdc.cen.uni-hamburg.de/thredds/dodsC/ftpthredds/hamtide/'


class HamtideResource:

    def __set__(self, obj, resource):
        if resource is None:
            resource = {'elevation': {}, 'velocity': {}}
            for constituent in obj.constituents:
                for key in resource.keys():
                    resource[key].update({constituent: None})
        else:
            raise NotImplementedError('Check that static files exist.')
            _resource = pathlib.Path(resource)
            for file in _resource.glob('*.nc'):
                print(file)

        obj.__dict__['resource'] = resource

    def __get__(self, obj, val):
        return obj.__dict__['resource']


class HAMTIDE(TidalDataProvider):
    ''' Wrapper for querying HAMTIDE model harmonic constituents.

    The source data can be either files from the static storage of on-the-fly
    querying from OpenDAP server (default).

    Reference:
        Taguchi, E., D. Stammer and W. Zahel (2010), Estimation of deep ocean
        tidal energy dissipation based on the high-resolution data-assimilative
        HAMTIDE model (to be submitted to J. Geophys. Res.).
    '''

    _resource = HamtideResource()

    def __init__(self, resource=None):
        self._resource = resource

    def get_elevation(self, constituent, vertices):
        amp = 0.01 * self._get_interpolation(
            'elevation', 'AMPL', constituent, vertices)
        phase = self._get_interpolation(
            'elevation', 'PHAS', constituent, vertices)
        return amp, phase

    def get_velocity(self, constituent, vertices):
        uamp = 0.01 * self._get_interpolation(
            'velocity', 'UAMP', constituent, vertices)
        uphase = self._get_interpolation(
            'velocity', 'UPHA', constituent, vertices)
        vamp = 0.01 * self._get_interpolation(
            'velocity', 'VAMP', constituent, vertices)
        vphase = self._get_interpolation(
            'velocity', 'VPHA', constituent, vertices)
        return uamp, uphase, vamp, vphase

    @property
    def constituents(self):
        return ['S2', 'Q1', 'P1', 'O1', 'N2', 'M2', 'K2', 'K1']

    @property
    def x(self):
        if not hasattr(self, '_x'):
            self._x = Dataset(base_url + 'k2.hamtide11a.nc')['LON'][:].data
        return self._x

    @property
    def y(self):
        if not hasattr(self, '_y'):
            self._y = Dataset(base_url + 'k2.hamtide11a.nc')['LAT'][:].data
        return self._y

    def _get_resource(self, variable, constituent) -> Dataset:
        resource = self._resource[variable][constituent]
        if resource is not None:
            return Dataset(resource)
        if variable == 'elevation':
            fname = f'{constituent.lower()}.hamtide11a.nc'
        if variable == 'velocity':
            fname = f'HAMcurrent11a_{constituent.lower()}.nc'
        return Dataset(base_url + fname)

    def _get_interpolation(self, phys_var, ncvar, constituent, vertices):
        xq = np.asarray(
            [x + 360. if x < 0. else x for x in vertices[:, 0]]).flatten()
        yq = vertices[:, 1].flatten()
        xidx = np.logical_and(self.x >= np.min(xq), self.x <= np.max(xq))
        yidx = np.logical_and(self.y >= np.min(yq), self.y <= np.max(yq))
        xi, yi = np.meshgrid(self.x[xidx], self.y[yidx])
        xi = xi.flatten()
        yi = yi.flatten()
        zi = self._get_resource(
                phys_var, constituent)[ncvar][yidx, xidx].flatten()
        return griddata(
            (xi[~zi.mask], yi[~zi.mask]),
            zi[~zi.mask],
            (xq, yq),
            method='nearest'
        )



from datetime import timedelta
import pathlib
from typing import Union, Dict, List, Any

from pyproj import Transformer, CRS  # type: ignore[import]
from shapely.geometry import (   # type: ignore[import]
    Polygon,
    MultiPolygon,
    Point
)

from pyschism.enums import (
    StationOutputVariables,
    StationOutputIndex
)


class Stations:

    def __init__(self, nspool_sta: Union[int, timedelta],
                 crs: Union[str, CRS] = None, elev: bool = False,
                 air_pressure: bool = False, windx: bool = False,
                 windy: bool = False, T: bool = False, S: bool = False,
                 u: bool = False, v: bool = False, w: bool = False):
        """Acts as container for requesting point outputs to SCHISM

        Args:
            nspool_sta: If type is <int>, this refers to output skip, else use
                a timedelta object.
            crs: Coordinate reference system for the container. All points
                added using the add_station() method should be in this CRS.
            elev (optional): Request elevation output, default: False
            air_pressure (optional): Request air_pressure output,
                default: False
            windx (optional): Request wind zonal output,
                default: False
            windy (optional): Request wind meridional output, default: False
            T (optional): Request temperature output, default: False
            S (optional): Request salinity output, default: False
            u (optional): Request zonal fluid velocity output, default: False
            v (optional): Request meridional fluid velocity output,
                default: False
            w (optional): Request vertical fluid velocity output,
                default: False
        """

        if not isinstance(nspool_sta, (int, timedelta)):
            raise TypeError('nspool_sta must be an int or timedelta')
        self.__nspool_sta = nspool_sta

        if isinstance(crs, str):
            crs = CRS.from_user_input(crs)
        self.__crs = crs

        # contains main station data (coordinates and id)
        self.__stations: List[Dict[str, Any]] = []

        # init the properties
        self.elev = elev
        self.air_pressure = air_pressure
        self.windx = windx
        self.windy = windy
        self.T = T
        self.S = S
        self.u = u
        self.v = v
        self.w = w

    def __iter__(self):
        for i, s in enumerate(self.__stations):
            yield i+1, s['x'], s['y'], s['z'], s['comment']

    def __str__(self):
        f = [f'{self.state}',
             f'{len(self.__stations)}']
        for i, x, y, z, comment in self:
            if comment is None:
                comment = ''
            else:
                comment = comment.split('!')
                comment = ' '.join(comment)
            f.append(f'{i} {x} {y} {z} ! {comment}')
        return '\n'.join(f)

    @staticmethod
    def from_file(file: Union[str, pathlib.Path],
                  nspool_sta: Union[int, timedelta],
                  crs: Union[str, CRS] = None, **kwargs):
        """Import station.in file as a Stations instance.

        Args:
            file: Path to the input.in file to import
            nspool_sta: Output frequency request
            crs (optional): Specify CRS of station.in. If no crs is given,
                the program assumes that the file is in the same CRS as the
                Mesh.
            elev (optional): Request elevation output, default: False
            air_pressure (optional): Request air_pressure output,
                default: False
            windx (optional): Request wind zonal output,
                default: False
            windy (optional): Request wind meridional output,
                default: False
            T (optional): Request temperature output, default: False
            S (optional): Request salinity output, default: False
            u (optional): Request zonal fluid velocity output,
                default: False
            v (optional): Request meridional fluid velocity output,
                default: False
            w (optional): Request vertical fluid velocity output,
                default: False

        Returns:
            instance of Stations
        """

        with open(pathlib.Path(file), 'r') as f:
            states = f.readline().split()[:9]
            # override file state with kwargs request
            for i, state in enumerate(states):
                states[i] = kwargs.get(
                    StationOutputIndex(i).name.lower(), False)
            stations = Stations(nspool_sta, crs=crs,
                                **{var.value: bool(states[i])
                                    for i, var in
                                    enumerate(StationOutputVariables)})
            comment: Union[str, None] = None
            for i in range(int(f.readline().split()[0])):
                line = f.readline().split()
                id, x, y, z = line[:4]
                if len(line) > 4:
                    comment = ' '.join(line[5:])
                else:
                    comment = None
                stations.add_station(float(x), float(y), float(z), comment)
        return stations

    def add_station(self, x: float, y: float, z: float = 0,
                    comment: str = None):
        """Adds a station coordinates to the container.

        Args:
            x: x-coordinate
            y: y-coordinate
            z: z-coordinate
            comment (optional): Can be used to store the COOPS id, useful
                during post-processing for validation.
        """
        self.__stations.append({'x': x, 'y': y, 'z': z, 'comment': comment})

    def get_active_vars(self) -> List[str]:
        """Returns a list of the names of activated output variables.

        Returns:
            list of strings for each active station output request.
        """
        return [var.value for var in StationOutputVariables
                if getattr(self, var.value) is True]

    def transform_to(self, dst_crs: Union[str, CRS]):
        """Transforms the horizontal coordinates of the stations.

        If :class:`pyschism.Stations` was instantiated with crs=None, then
        this function cannot be used.
        """
        if isinstance(dst_crs, str):
            dst_crs = CRS.from_user_input(dst_crs)
        else:
            if not isinstance(dst_crs, CRS):
                raise TypeError(f'Input must be {str} or {CRS}')
        if dst_crs.equals(self.crs):
            return
        transformer = Transformer.from_crs(self.crs, dst_crs, always_xy=True)
        x = [_['x'] for _ in self.__stations]
        y = [_['y'] for _ in self.__stations]
        xy = list(zip(*transformer.transform(x, y)))
        for i, (x, y) in enumerate(xy):
            self.__stations[i]['x'] = x
            self.__stations[i]['y'] = y
        self.__crs = dst_crs

    def clip(self, geometry: Union[Polygon, MultiPolygon]):
        """Eliminates any stations not contained within the given geometry.

        It is assumed that the input geometry matches the CRS of the
        Stations instance

        Args:
            geometry: Polygon or MultiPolygon used for clipping the
                stations. Normally this comes from Hgrid.get_multipolygon()
        """

        eliminate: List[int] = []
        for id, x, y, _, _ in self:
            if not geometry.contains(Point(x, y)):
                eliminate.insert(id-1, 0)
        for id in eliminate:
            del(self.__stations[id])

    def write(self, path: Union[str, pathlib.Path], overwrite: bool = False):
        """Writes the SCHISM station.in file to disk.

        Args:
            path: Path to output on disk (need to include filename).
            overwrite (optional): Allow/disallow file overwrite.
        """
        path = pathlib.Path(path)
        if path.is_file() and not overwrite:
            raise IOError("File exists and overwrite is False.")
        with open(path, 'w') as f:
            f.write(str(self))

    @property
    def stations(self) -> List[Dict]:
        """Returns list of currently loaded stations."""
        return self.__stations

    @property
    def nspool_sta(self) -> Union[int, timedelta]:
        """Returns output frequency."""
        return self.__nspool_sta

    @property
    def crs(self) -> Union[CRS, None]:
        """Returns coordinate reference system of the current instance."""
        return self.__crs

    @property
    def state(self) -> str:
        """Returns corresponding string that goes into bctide.in"""
        return ' '.join([str(int(getattr(self, var.value)))
                         for var in StationOutputVariables])

    @property
    def elev(self) -> bool:
        """Returns state (ON/OFF) of variable elev request."""
        return self.__elev

    @elev.setter
    def elev(self, elev: bool):
        assert isinstance(elev, bool), 'Argument to "elev" must be boolean'
        self.__elev = elev

    @property
    def air_pressure(self) -> bool:
        """Returns state (ON/OFF) of variable air_pressure request."""
        return self.__air_pressure

    @air_pressure.setter
    def air_pressure(self, air_pressure: bool):
        assert isinstance(air_pressure, bool), \
            'Argument to "air_pressure" must be boolean'
        self.__air_pressure = air_pressure

    @property
    def windx(self) -> bool:
        """Returns state (ON/OFF) of variable windx request."""
        return self.__windx

    @windx.setter
    def windx(self, windx: bool):
        assert isinstance(windx, bool), 'Argument to "windx" must be boolean'
        self.__windx = windx

    @property
    def windy(self) -> bool:
        """Returns state (ON/OFF) of variable windy request."""
        return self.__windy

    @windy.setter
    def windy(self, windy: bool):
        assert isinstance(windy, bool), 'Argument to "windy" must be boolean'
        self.__windy = windy

    @property
    def T(self) -> bool:
        """Returns state (ON/OFF) of variable T request."""
        return self.__T

    @T.setter
    def T(self, T: bool):
        assert isinstance(T, bool), 'Argument to "T" must be boolean'
        self.__T = T

    @property
    def S(self) -> bool:
        """Returns state (ON/OFF) of variable S request."""
        return self.__S

    @S.setter
    def S(self, S: bool):
        assert isinstance(S, bool), 'Argument to "S" must be boolean'
        self.__S = S

    @property
    def u(self) -> bool:
        """Returns state (ON/OFF) of variable u request."""
        return self.__u

    @u.setter
    def u(self, u: bool):
        assert isinstance(u, bool), 'Argument to "u" must be boolean'
        self.__u = u

    @property
    def v(self) -> bool:
        """Returns state (ON/OFF) of variable v request."""
        return self.__v

    @v.setter
    def v(self, v: bool):
        assert isinstance(v, bool), 'Argument to "v" must be boolean'
        self.__v = v

    @property
    def w(self) -> bool:
        """Returns state (ON/OFF) of variable w request."""
        return self.__w

    @w.setter
    def w(self, w: bool):
        assert isinstance(w, bool), 'Argument to "w" must be boolean'
        self.__w = w
